package multex.tool;

import java.io.FileWriter;
import java.io.PrintWriter;

import multex.Failure;

import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.RootDoc;

/**This doclet collects the exception message text pattern of each concrete Throwable-subclass.
 * Scans all sources, takes the javadoc main comment text of each concrete subclass of Throwable, and appends all of them to the 
 * file given by the Doclet option -out
*/ 
public class ExceptionMessagesDoclet {
	
	
	private static final String myClassName = "ExceptionMessagesDoclet";

	/**Qualified class name of java.lang.Object*/
	private static final String JAVA_LANG_OBJECT = Object.class.getName();
	
	/**Qualified class name of java.lang.Throwable*/
	private static final String JAVA_LANG_THROWABLE = Throwable.class.getName();
	
	/**The name of the file, where to append the rule violation messages.*/
	private static String outFileName;
	
	/**The destination, where to write the rule violations.*/
	private static java.io.PrintWriter out;
	
	/**Check for doclet-added options. 
	 * Only one option is allowed her: -out filename
	 * 
	 * @param option the name of the option including the minus character, for example "-out".
	 * @return number of arguments on the command line for an option including the option name itself. Zero return means option not known. Negative value means error occurred.
	 */
	public static int optionLength(final String option){
		if("-out".equals(option)){return 2;}
		return 0;
	}
	
	/**Entry point to execute this Doclet. It will be invoked by the Javadoc tool.*/ 
    public static boolean start(final RootDoc i_rootDoc) throws Exception {
    	new ExceptionMessagesDoclet()._execute(i_rootDoc);
        System.out.println("\nAll exception message text patterns have been appended to file " + outFileName);
    	return true;
    }

	/**Appends all exception message text patterns to file.
	 * @param rootDoc Comprising all classes to visit.
	 * @throws Exception An error occured inside of this doclet.
	 */
	private void _execute(final RootDoc rootDoc) throws Exception {
		outFileName = _getOutputFilename(rootDoc.options());
    	out = new PrintWriter(new FileWriter(outFileName, /*append*/true));
        _visitAll(rootDoc);
        out.close();
	}

    /**Gets the value of the doclet option -overview.
     * This will be used as the name of the file to append the messages to.
     * Unfortunately the ANT Javadoc task does not permit to pass user specified options,
     * so we have to misuse this predefined one.
     * @param options Each element of it is an array with it's first element being the option's name, the following elements being option values.
     * @return the filename where to add exception text property definitions
     */
	private String _getOutputFilename(final String[][] options) {
		for(int i=0; i<options.length; i++){
			final String[] option = options[i];
			if("-overview".equals(option[0])){
				return option[1];
			}
			if("-out".equals(option[0])){
				return option[1];
			}
		}
		throw new Failure("Missing name of output file, where to append the extracted exception text property definitions.\nUse Javadoc option -overview or -out to indicate it!", (Throwable)null);
	}

	/**Visits all classes contained in rootDoc. Visits the nested classes, too.*/
	private void _visitAll(final RootDoc rootDoc) throws Exception {
		final ClassDoc[] classDocs = rootDoc.classes();
        for (int i = 0; i < classDocs.length; ++i) {
            final ClassDoc classDoc = classDocs[i];
            //System.out.println("Visiting class " + classDoc);
            _visitClass(classDoc);
        }
	}

	/**Vists the class in classDoc.
	 * @throws TooNestedThrowableException */
	private void _visitClass(final ClassDoc classDoc) throws TooNestedThrowableException{
		_printMessageTextPatternOfThrowable(classDoc);
		/*Rekursion anscheinend unntig!
		final ClassDoc[] innerClasses = classDoc.innerClasses();
		for(int i = 0; i < innerClasses.length; ++i){
		    final ClassDoc innerClassDoc = innerClasses[i];
		    _visitClass(innerClassDoc);
		}
		*/
	}

	/**Prints the message text pattern for the class, if it is a concrete subclass of Throwable.
	 * The main commentText (without tags) of the class is written in the Java property file format to {@link #out}.
	 * 
	 * @param classDoc representing the class to be checked.
	 * @throws TooNestedThrowableException 
	 */
	private void _printMessageTextPatternOfThrowable(final ClassDoc classDoc) throws TooNestedThrowableException {
		if(classDoc.isAbstract() || classDoc.isOrdinaryClass()){return;}
		//Now classdoc is a concrete interface, annotation type, enum, exception, or error
		//out.println("class: " + classDoc);
		if(!_isSubclassOfThrowable(classDoc)){return;}
		final String commentText = classDoc.commentText();
		if(commentText.length()<10){
			System.err.println("Too short (<10ch) message text \"" + commentText + "\" for Throwable " + classDoc);
			return;
		}
		final String key = _internalClassName(classDoc);
		out.print(key);
		out.print(" = ");
		out.print(commentText);
		out.println();
	}

	/**Throwable class {canonicalName} has more than one nesting level. Handling not yet implemented.*/
	public static final class TooNestedThrowableException extends Exception {
		public TooNestedThrowableException(final String canonicalName){
			super("Throwable class " + canonicalName + " has more than one nesting level. Handling not yet implemented.");
		}
	}

	/**Returns the internal class name of the class.
	 * @param classDoc the class documentation, from which to get the internal name.
	 * @return the class name in the format of java.lang.Class.getName(), 
	 *         that means nested class identifiers are separated by a dollar sign ($), 
	 *         rather than by a period (.) from the identifier of the containing class. 
	 * @throws TooNestedThrowableException The internal class name would have more than one dollar character.
	 */
	private String _internalClassName(final ClassDoc classDoc) throws TooNestedThrowableException {
		final String canonicalName = classDoc.toString();
		final ClassDoc containingClass = classDoc.containingClass();
		if(null==containingClass){
			//canonicalName does not contain a dollar ($) character.
			return canonicalName;
		}
		if(null!=containingClass.containingClass()){
			throw new TooNestedThrowableException(canonicalName);
		}
		final int indexOfLastDollar = canonicalName.lastIndexOf('.');
		final char[] characters = canonicalName.toCharArray();
		characters[indexOfLastDollar] = '$';
		final String result = new String(characters);
		return result;
	}

	private boolean _isSubclassOfThrowable(final ClassDoc classDoc){
		final ClassDoc superClass = classDoc.superclass();
		final String superClassName = superClass.toString();
		//out.println("super: " + superClass);
		if(JAVA_LANG_THROWABLE.equals(superClassName)){return true;}
		if(JAVA_LANG_OBJECT.equals(superClassName)){return false;}
		return _isSubclassOfThrowable(superClass);
	}


}
